@@ -57,11 +57,13 @@ struct pl_context *pl_context_create(int api_ver,
     if (api_ver != PL_API_VER) {
         fprintf(stderr,
                "*************************************************************\n"
-               "libplacebo: ABI mismatch detected!\n\n"
+               "libplacebo: ABI mismatch detected! (requested: %d, compiled: %d)\n"
+               "\n"
                "This is usually indicative of a linking mismatch, and will\n"
                "result in serious issues including stack corruption, random\n"
                "crashes and arbitrary code execution. Aborting as a safety\n"
-               "precaution. Fix your system!\n");
+               "precaution. Fix your system!\n",
+               api_ver, PL_API_VER);
         abort();
     }
 
@@ -73,6 +75,7 @@ struct pl_context *pl_context_create(int api_ver,
 
     struct pl_context *ctx = talloc_zero(NULL, struct pl_context);
     ctx->params = *PL_DEF(params, &pl_context_default_params);
+    pl_info(ctx, "Initialized libplacebo %s (API v%d)", PL_VERSION, PL_API_VER);
     return ctx;
 }
 
@@ -113,6 +116,8 @@ void pl_log_simple(void *stream, enum pl_log_level level, const char *msg)
 
     FILE *h = default_stream(stream, level);
     fprintf(h, "%5s: %s\n", prefix[level], msg);
+    if (level <= PL_LOG_WARN)
+        fflush(h);
 }
 
 void pl_log_color(void *stream, enum pl_log_level level, const char *msg)
@@ -128,6 +133,8 @@ void pl_log_color(void *stream, enum pl_log_level level, const char *msg)
 
     FILE *h = default_stream(stream, level);
     fprintf(h, "\033[%sm%s\033[0m\n", color[level], msg);
+    if (level <= PL_LOG_WARN)
+        fflush(h);
 }
 
 void pl_msg(struct pl_context *ctx, enum pl_log_level lev, const char *fmt, ...)
