@@ -217,15 +217,29 @@ static void add_var(struct pl_dispatch *dp, struct bstr *body,
     }
 }
 
-static void add_buffer_vars(struct pl_dispatch *dp, struct bstr *body,
-                            const struct pl_buffer_var *vars, int num)
+static int cmp_buffer_var(const void *pa, const void *pb)
 {
+    const struct pl_buffer_var * const *a = pa, * const *b = pb;
+    return PL_CMP((*a)->layout.offset, (*b)->layout.offset);
+}
+
+static void add_buffer_vars(struct pl_dispatch *dp, struct bstr *body,
+                            const struct pl_buffer_var *vars, int num,
+                            void *tmp)
+{
+    // Sort buffer vars
+    const struct pl_buffer_var **sorted_vars = NULL;
+    TARRAY_RESIZE(tmp, sorted_vars, num);
+    for (int i = 0; i < num; i++)
+        sorted_vars[i] = &vars[i];
+    qsort(sorted_vars, num, sizeof(sorted_vars[0]), cmp_buffer_var);
+
     ADD(body, "{\n");
     for (int i = 0; i < num; i++) {
         // Add an explicit offset wherever possible
         if (dp->gpu->glsl.version >= 440)
-            ADD(body, "    layout(offset=%zu) ", vars[i].layout.offset);
-        add_var(dp, body, &vars[i].var);
+            ADD(body, "    layout(offset=%zu) ", sorted_vars[i]->layout.offset);
+        add_var(dp, body, &sorted_vars[i]->var);
     }
     ADD(body, "};\n");
 }
@@ -242,7 +256,8 @@ static ident_t sh_var_from_va(struct pl_shader *sh, const char *name,
 
 static void generate_shaders(struct pl_dispatch *dp, struct pass *pass,
                              struct pl_pass_params *params,
-                             struct pl_shader *sh, ident_t vert_pos)
+                             struct pl_shader *sh, ident_t vert_pos,
+                             void *tmp)
 {
     const struct pl_gpu *gpu = dp->gpu;
     const struct pl_shader_res *res = pl_shader_finalize(sh);
@@ -254,14 +269,24 @@ static void generate_shaders(struct pl_dispatch *dp, struct pass *pass,
         ADD(pre, "#extension GL_ARB_compute_shader : enable\n");
 
     // Enable all extensions needed for different types of input
-    bool has_ssbo = false, has_ubo = false, has_img = false, has_texel = false;
+    bool has_ssbo = false, has_ubo = false, has_img = false, has_texel = false,
+         has_ext = false;
     for (int i = 0; i < res->num_descriptors; i++) {
-        switch (params->descriptors[i].type) {
+        switch (res->descriptors[i].desc.type) {
         case PL_DESC_STORAGE_IMG: has_img = true; break;
         case PL_DESC_BUF_UNIFORM: has_ubo = true; break;
         case PL_DESC_BUF_STORAGE: has_ssbo = true; break;
         case PL_DESC_BUF_TEXEL_UNIFORM: // fall through
         case PL_DESC_BUF_TEXEL_STORAGE: has_texel = true; break;
+        case PL_DESC_SAMPLED_TEX: {
+            const struct pl_tex *tex = res->descriptors[i].object;
+            switch (tex->sampler_type) {
+            case PL_SAMPLER_NORMAL: break;
+            case PL_SAMPLER_RECT: break;
+            case PL_SAMPLER_EXTERNAL: has_ext = true; break;
+            }
+            break;
+        }
         default: break;
         }
     }
@@ -274,6 +299,8 @@ static void generate_shaders(struct pl_dispatch *dp, struct pass *pass,
         ADD(pre, "#extension GL_ARB_shader_storage_buffer_object : enable\n");
     if (has_texel)
         ADD(pre, "#extension GL_ARB_texture_buffer_object : enable\n");
+    if (has_ext)
+        ADD(pre, "#extension GL_OES_EGL_image_external : enable\n");
 
     if (gpu->glsl.gles) {
         ADD(pre, "precision mediump float;\n");
@@ -350,18 +377,24 @@ static void generate_shaders(struct pl_dispatch *dp, struct pass *pass,
 
     // Add all of the push constants as their own element
     if (params->push_constants_size) {
-        ADD(glsl, "layout(std430, push_constant) uniform PushC {\n");
+        // We re-use add_buffer_vars to make sure variables are sorted, this
+        // is important because the push constants can be out-of-order in
+        // `pass->vars`
+        struct pl_buffer_var *pc_bvars = NULL;
+        int num_bvars = 0;
+
         for (int i = 0; i < res->num_variables; i++) {
-            struct pl_var *var = &res->variables[i].var;
-            struct pass_var *pv = &pass->vars[i];
-            if (pv->type != PASS_VAR_PUSHC)
+            if (pass->vars[i].type != PASS_VAR_PUSHC)
                 continue;
-            // Note: Don't remove this offset, since the push constants
-            // can be out-of-order in `pass->vars`!
-            ADD(glsl, "    layout(offset=%zu) ", pv->layout.offset);
-            add_var(dp, glsl, var);
+
+            TARRAY_APPEND(tmp, pc_bvars, num_bvars, (struct pl_buffer_var) {
+                .var = res->variables[i].var,
+                .layout = pass->vars[i].layout,
+            });
         }
-        ADD(glsl, "};\n");
+
+        ADD(glsl, "layout(std430, push_constant) uniform PushC ");
+        add_buffer_vars(dp, glsl, pc_bvars, num_bvars, tmp);
     }
 
     // Add all of the required descriptors
@@ -371,20 +404,41 @@ static void generate_shaders(struct pl_dispatch *dp, struct pass *pass,
 
         switch (desc->type) {
         case PL_DESC_SAMPLED_TEX: {
-            static const char *types[] = {
-                [1] = "sampler1D",
-                [2] = "sampler2D",
-                [3] = "sampler3D",
+            static const char *types[][4] = {
+                [PL_SAMPLER_NORMAL][1]  = "sampler1D",
+                [PL_SAMPLER_NORMAL][2]  = "sampler2D",
+                [PL_SAMPLER_NORMAL][3]  = "sampler3D",
+                [PL_SAMPLER_RECT][2]    = "sampler2DRect",
+                [PL_SAMPLER_EXTERNAL][2] = "samplerExternalOES",
             };
 
+            const struct pl_tex *tex = sd->object;
+            int dims = pl_tex_params_dimension(tex->params);
+            const char *type = types[tex->sampler_type][dims];
+            pl_assert(type);
+
+            static const char prefixes[PL_FMT_TYPE_COUNT] = {
+                [PL_FMT_FLOAT]  = ' ',
+                [PL_FMT_UNORM]  = ' ',
+                [PL_FMT_SNORM]  = ' ',
+                [PL_FMT_UINT]   = 'u',
+                [PL_FMT_SINT]   = 'i',
+            };
+
+            char prefix = prefixes[tex->params.format->type];
+            pl_assert(prefix);
+
+            const char *prec = "";
+            if (prefix != ' ' && gpu->glsl.gles)
+                prec = "highp ";
+
             // Vulkan requires explicit bindings; GL always sets the
             // bindings manually to avoid relying on the user doing so
             if (gpu->glsl.vulkan)
                 ADD(glsl, "layout(binding=%d) ", desc->binding);
 
-            const struct pl_tex *tex = sd->object;
-            int dims = pl_tex_params_dimension(tex->params);
-            ADD(glsl, "uniform %s %s;\n", types[dims], desc->name);
+            pl_assert(type && prefix);
+            ADD(glsl, "uniform %s%c%s %s;\n", prec, prefix, type, desc->name);
             break;
         }
 
@@ -408,7 +462,7 @@ static void generate_shaders(struct pl_dispatch *dp, struct pass *pass,
             } else if (gpu->glsl.version >= 130) {
                 ADD(glsl, "layout(%s) ", format);
             }
-            ADD(glsl, "%s uniform %s %s;\n", access, types[dims], desc->name);
+            ADD(glsl, "%s restrict uniform %s %s;\n", access, types[dims], desc->name);
             break;
         }
 
@@ -419,7 +473,7 @@ static void generate_shaders(struct pl_dispatch *dp, struct pass *pass,
                 ADD(glsl, "layout(std140) ");
             }
             ADD(glsl, "uniform %s ", desc->name);
-            add_buffer_vars(dp, glsl, sd->buffer_vars, sd->num_buffer_vars);
+            add_buffer_vars(dp, glsl, sd->buffer_vars, sd->num_buffer_vars, tmp);
             break;
 
         case PL_DESC_BUF_STORAGE:
@@ -428,9 +482,11 @@ static void generate_shaders(struct pl_dispatch *dp, struct pass *pass,
             } else if (gpu->glsl.version >= 140) {
                 ADD(glsl, "layout(std430) ");
             }
-            ADD(glsl, "%s buffer %s ", pl_desc_access_glsl_name(desc->access),
-                desc->name);
-            add_buffer_vars(dp, glsl, sd->buffer_vars, sd->num_buffer_vars);
+            // XXX: back-ported bugfix, hard-code all storage buffers as
+            // coherent to avoid breaking API
+            ADD(glsl, "%s coherent restrict buffer %s ",
+                pl_desc_access_glsl_name(desc->access), desc->name);
+            add_buffer_vars(dp, glsl, sd->buffer_vars, sd->num_buffer_vars, tmp);
             break;
 
         case PL_DESC_BUF_TEXEL_UNIFORM:
@@ -448,7 +504,7 @@ static void generate_shaders(struct pl_dispatch *dp, struct pass *pass,
             } else {
                 ADD(glsl, "layout(%s) ", format);
             }
-            ADD(glsl, "%s uniform imageBuffer %s;\n", access, desc->name);
+            ADD(glsl, "%s restrict uniform imageBuffer %s;\n", access, desc->name);
             break;
         }
         default: abort();
@@ -503,26 +559,31 @@ static bool blend_equal(const struct pl_blend_params *a,
 
 static struct pass *find_pass(struct pl_dispatch *dp, struct pl_shader *sh,
                               const struct pl_tex *target, ident_t vert_pos,
-                              const struct pl_blend_params *blend)
+                              const struct pl_blend_params *blend, bool load)
 {
     uint64_t sig = pl_shader_signature(sh);
 
     for (int i = 0; i < dp->num_passes; i++) {
-        const struct pass *p = dp->passes[i];
+        struct pass *p = dp->passes[i];
         if (p->signature != sig)
             continue;
 
+        // Failed shader, no additional checks needed
+        if (!p->pass)
+            return p;
+
         if (pl_shader_is_compute(sh)) {
             // no special requirements besides the signature
-            return dp->passes[i];
+            return p;
         } else {
             pl_assert(target);
             const struct pl_fmt *tfmt;
             tfmt = p->pass->params.target_dummy.params.format;
             bool raster_ok = target->params.format == tfmt;
             raster_ok &= blend_equal(p->pass->params.blend_params, blend);
+            raster_ok &= load == p->pass->params.load_target;
             if (raster_ok)
-                return dp->passes[i];
+                return p;
         }
     }
 
@@ -530,7 +591,6 @@ static struct pass *find_pass(struct pl_dispatch *dp, struct pl_shader *sh,
 
     struct pass *pass = talloc_zero(dp, struct pass);
     pass->signature = sig;
-    pass->failed = true; // will be set to false on success
     pass->ubo_desc = (struct pl_shader_desc) {
         .desc = {
             .name = "UBO",
@@ -550,6 +610,7 @@ static struct pass *find_pass(struct pl_dispatch *dp, struct pl_shader *sh,
     if (params.type == PL_PASS_RASTER) {
         assert(target);
         params.target_dummy = *target;
+        params.load_target = load;
 
         // Fill in the vertex attributes array
         params.num_vertex_attribs = res->num_vertex_attribs;
@@ -559,7 +620,7 @@ static struct pass *find_pass(struct pl_dispatch *dp, struct pl_shader *sh,
         int va_loc = 0;
         for (int i = 0; i < res->num_vertex_attribs; i++) {
             struct pl_vertex_attrib *va = &params.vertex_attribs[i];
-            *va = res->vertex_attribs[i].attr;
+            *va = sh->vertex_attribs[i].attr;
 
             // Mangle the name to make sure it doesn't conflict with the
             // fragment shader input
@@ -590,12 +651,12 @@ static struct pass *find_pass(struct pl_dispatch *dp, struct pl_shader *sh,
     // want inside PCs, and then a second time to opportunistically place the rest.
     pass->vars = talloc_zero_array(pass, struct pass_var, res->num_variables);
     for (int i = 0; i < res->num_variables; i++) {
-        if (!add_pass_var(dp, tmp, pass, &params, &res->variables[i], &pass->vars[i], false))
+        if (!add_pass_var(dp, tmp, pass, &params, &sh->variables[i], &pass->vars[i], false))
             goto error;
     }
 
     for (int i = 0; i < res->num_variables; i++) {
-        if (!add_pass_var(dp, tmp, pass, &params, &res->variables[i], &pass->vars[i], true))
+        if (!add_pass_var(dp, tmp, pass, &params, &sh->variables[i], &pass->vars[i], true))
             goto error;
     }
 
@@ -627,7 +688,7 @@ static struct pass *find_pass(struct pl_dispatch *dp, struct pl_shader *sh,
     rparams->desc_bindings = talloc_zero_array(pass, struct pl_desc_binding, num);
     for (int i = 0; i < num; i++) {
         struct pl_desc *desc = &params.descriptors[i];
-        *desc = res->descriptors[i].desc;
+        *desc = sh->descriptors[i].desc;
         desc->binding = binding[pl_desc_namespace(dp->gpu, desc->type)]++;
     }
 
@@ -642,15 +703,14 @@ static struct pass *find_pass(struct pl_dispatch *dp, struct pl_shader *sh,
     rparams->push_constants = talloc_zero_size(pass, params.push_constants_size);
 
     // Finally, finalize the shaders and create the pass itself
-    generate_shaders(dp, pass, &params, sh, vert_pos);
+    generate_shaders(dp, pass, &params, sh, vert_pos, tmp);
     pass->pass = rparams->pass = pl_pass_create(dp->gpu, &params);
     if (!pass->pass) {
         PL_ERR(dp, "Failed creating render pass for dispatch");
         goto error;
     }
 
-    pass->failed = false;
-
+    // fall through
 error:
     pass->ubo_desc = (struct pl_shader_desc) {0}; // contains temporary pointers
     talloc_free(tmp);
@@ -702,15 +762,11 @@ static void update_pass_var(struct pl_dispatch *dp, struct pass *pass,
     };
 }
 
-static void translate_compute_shader(struct pl_dispatch *dp,
-                                     struct pl_shader *sh,
-                                     const struct pl_tex *target,
-                                     const struct pl_rect2d *rc,
-                                     const struct pl_blend_params *blend)
+static void compute_vertex_attribs(struct pl_dispatch *dp, struct pl_shader *sh,
+                                   int width, int height, ident_t *out_scale)
 {
     // Simulate vertex attributes using global definitions
-    int width = abs(pl_rect_w(*rc)), height = abs(pl_rect_h(*rc));
-    ident_t out_scale = sh_var(sh, (struct pl_shader_var) {
+    *out_scale = sh_var(sh, (struct pl_shader_var) {
         .var     = pl_var_vec2("out_scale"),
         .data    = &(float[2]){ 1.0 / width, 1.0 / height },
         .dynamic = true,
@@ -719,10 +775,10 @@ static void translate_compute_shader(struct pl_dispatch *dp,
     GLSLP("#define frag_pos(id) (vec2(id) + vec2(0.5)) \n"
           "#define frag_map(id) (%s * frag_pos(id))    \n"
           "#define gl_FragCoord vec4(frag_pos(gl_GlobalInvocationID), 0.0, 1.0) \n",
-          out_scale);
+          *out_scale);
 
     for (int n = 0; n < sh->res.num_vertex_attribs; n++) {
-        const struct pl_shader_va *sva = &sh->res.vertex_attribs[n];
+        const struct pl_shader_va *sva = &sh->vertex_attribs[n];
 
         ident_t points[4];
         for (int i = 0; i < PL_ARRAY_SIZE(points); i++) {
@@ -740,18 +796,28 @@ static void translate_compute_shader(struct pl_dispatch *dp,
              points[0], points[1], points[2], points[3],
              sva->attr.name, sva->attr.name);
     }
+}
+
+static void translate_compute_shader(struct pl_dispatch *dp,
+                                     struct pl_shader *sh,
+                                     const struct pl_rect2d *rc,
+                                     const struct pl_dispatch_params *params)
+{
+    int width = abs(pl_rect_w(*rc)), height = abs(pl_rect_h(*rc));
+    ident_t out_scale;
+    compute_vertex_attribs(dp, sh, width, height, &out_scale);
 
     // Simulate a framebuffer using storage images
-    pl_assert(target->params.storable);
+    pl_assert(params->target->params.storable);
     pl_assert(sh->res.output == PL_SHADER_SIG_COLOR);
     ident_t fbo = sh_desc(sh, (struct pl_shader_desc) {
         .desc = {
             .name    = "out_image",
             .type    = PL_DESC_STORAGE_IMG,
-            .access  = blend ? PL_DESC_ACCESS_READWRITE
-                             : PL_DESC_ACCESS_WRITEONLY,
+            .access  = params->blend_params ? PL_DESC_ACCESS_READWRITE
+                                            : PL_DESC_ACCESS_WRITEONLY,
         },
-        .object = target,
+        .object = params->target,
     });
 
     ident_t base = sh_var(sh, (struct pl_shader_var) {
@@ -771,7 +837,7 @@ static void translate_compute_shader(struct pl_dispatch *dp,
     GLSL("ivec2 pos = %s + dir * ivec2(gl_GlobalInvocationID);\n", base);
     GLSL("vec2 fpos = %s * vec2(gl_GlobalInvocationID);\n", out_scale);
     GLSL("if (max(fpos.x, fpos.y) < 1.0) {\n");
-    if (blend) {
+    if (params->blend_params) {
         GLSL("vec4 orig = imageLoad(%s, pos);\n", fbo);
 
         static const char *modes[] = {
@@ -783,19 +849,19 @@ static void translate_compute_shader(struct pl_dispatch *dp,
 
         GLSL("color = vec4(color.rgb * vec3(%s), color.a * %s) \n"
              "      + vec4(orig.rgb  * vec3(%s), orig.a  * %s);\n",
-             modes[blend->src_rgb], modes[blend->src_alpha],
-             modes[blend->dst_rgb], modes[blend->dst_alpha]);
+             modes[params->blend_params->src_rgb],
+             modes[params->blend_params->src_alpha],
+             modes[params->blend_params->dst_rgb],
+             modes[params->blend_params->dst_alpha]);
     }
     GLSL("imageStore(%s, pos, color);\n", fbo);
     GLSL("}\n");
     sh->res.output = PL_SHADER_SIG_NONE;
 }
 
-bool pl_dispatch_finish(struct pl_dispatch *dp, struct pl_shader **psh,
-                        const struct pl_tex *target, const struct pl_rect2d *rc,
-                        const struct pl_blend_params *blend)
+bool pl_dispatch_finish(struct pl_dispatch *dp, const struct pl_dispatch_params *params)
 {
-    struct pl_shader *sh = *psh;
+    struct pl_shader *sh = *params->shader;
     const struct pl_shader_res *res = &sh->res;
     bool ret = false;
 
@@ -814,7 +880,7 @@ bool pl_dispatch_finish(struct pl_dispatch *dp, struct pl_shader **psh,
         goto error;
     }
 
-    const struct pl_tex_params *tpars = &target->params;
+    const struct pl_tex_params *tpars = &params->target->params;
     if (pl_tex_params_dimension(*tpars) != 2 || !tpars->renderable) {
         PL_ERR(dp, "Trying to dispatch a shader using an invalid target "
                "texture. The target must be a renderable 2D texture.");
@@ -825,12 +891,22 @@ bool pl_dispatch_finish(struct pl_dispatch *dp, struct pl_shader **psh,
         PL_ERR(dp, "Trying to dispatch using a compute shader with a "
                "non-storable target texture.");
         goto error;
+    } else if (tpars->storable && (dp->gpu->caps & PL_GPU_CAP_PARALLEL_COMPUTE)) {
+        if (sh_try_compute(sh, 16, 16, true, 0))
+            PL_TRACE(dp, "Upgrading fragment shader to compute shader.");
     }
 
-    struct pl_rect2d full = {0, 0, tpars->w, tpars->h};
-    rc = PL_DEF(rc, &full);
+    struct pl_rect2d rc = params->rect;
+    if (!pl_rect_w(rc)) {
+        rc.x0 = 0;
+        rc.x1 = tpars->w;
+    }
+    if (!pl_rect_h(rc)) {
+        rc.y0 = 0;
+        rc.y1 = tpars->h;
+    }
 
-    int w, h, tw = abs(pl_rect_w(*rc)), th = abs(pl_rect_h(*rc));
+    int w, h, tw = abs(pl_rect_w(rc)), th = abs(pl_rect_h(rc));
     if (pl_shader_output_size(sh, &w, &h) && (w != tw || h != th))
     {
         PL_ERR(dp, "Trying to dispatch a shader with explicit output size "
@@ -843,40 +919,52 @@ bool pl_dispatch_finish(struct pl_dispatch *dp, struct pl_shader **psh,
 
     if (pl_shader_is_compute(sh)) {
         // Translate the compute shader to simulate vertices etc.
-        translate_compute_shader(dp, sh, target, rc, blend);
+        translate_compute_shader(dp, sh, &rc, params);
     } else {
         // Add the vertex information encoding the position
         vert_pos = sh_attr_vec2(sh, "position", &(const struct pl_rect2df) {
-            .x0 = 2.0 * rc->x0 / tpars->w - 1.0,
-            .y0 = 2.0 * rc->y0 / tpars->h - 1.0,
-            .x1 = 2.0 * rc->x1 / tpars->w - 1.0,
-            .y1 = 2.0 * rc->y1 / tpars->h - 1.0,
+            .x0 = 2.0 * rc.x0 / tpars->w - 1.0,
+            .y0 = 2.0 * rc.y0 / tpars->h - 1.0,
+            .x1 = 2.0 * rc.x1 / tpars->w - 1.0,
+            .y1 = 2.0 * rc.y1 / tpars->h - 1.0,
         });
     }
 
-    struct pass *pass = find_pass(dp, sh, target, vert_pos, blend);
+    // We need to set pl_pass_params.load_target when either blending is
+    // enabled or we're drawing to some scissored sub-rect of the texture
+    struct pl_rect2d full = { 0, 0, tpars->w, tpars->h };
+    struct pl_rect2d rc_norm = rc;
+    pl_rect2d_normalize(&rc_norm);
+    rc_norm.x0 = PL_MAX(rc_norm.x0, 0);
+    rc_norm.y0 = PL_MAX(rc_norm.y0, 0);
+    rc_norm.x1 = PL_MIN(rc_norm.x1, tpars->w);
+    rc_norm.y1 = PL_MIN(rc_norm.y1, tpars->h);
+    bool load = params->blend_params || !pl_rect2d_eq(rc_norm, full);
+
+    struct pass *pass = find_pass(dp, sh, params->target, vert_pos,
+                                  params->blend_params, load);
 
     // Silently return on failed passes
-    if (pass->failed)
+    if (!pass->pass)
         goto error;
 
     struct pl_pass_run_params *rparams = &pass->run_params;
 
     // Update the descriptor bindings
     for (int i = 0; i < sh->res.num_descriptors; i++)
-        rparams->desc_bindings[i].object = sh->res.descriptors[i].object;
+        rparams->desc_bindings[i].object = sh->descriptors[i].object;
 
     // Update all of the variables (if needed)
     rparams->num_var_updates = 0;
     for (int i = 0; i < res->num_variables; i++)
-        update_pass_var(dp, pass, &res->variables[i], &pass->vars[i]);
+        update_pass_var(dp, pass, &sh->variables[i], &pass->vars[i]);
 
     // Update the vertex data
     if (rparams->vertex_data) {
         uintptr_t vert_base = (uintptr_t) rparams->vertex_data;
         size_t stride = rparams->pass->params.vertex_stride;
         for (int i = 0; i < res->num_vertex_attribs; i++) {
-            struct pl_shader_va *sva = &res->vertex_attribs[i];
+            const struct pl_shader_va *sva = &sh->vertex_attribs[i];
             struct pl_vertex_attrib *va = &rparams->pass->params.vertex_attribs[i];
 
             size_t size = sva->attr.fmt->texel_size;
@@ -889,8 +977,8 @@ bool pl_dispatch_finish(struct pl_dispatch *dp, struct pl_shader **psh,
     // For compute shaders: also update the dispatch dimensions
     if (pl_shader_is_compute(sh)) {
         // Round up to make sure we don-t leave off a part of the target
-        int width = abs(pl_rect_w(*rc)),
-            height = abs(pl_rect_h(*rc)),
+        int width = abs(pl_rect_w(rc)),
+            height = abs(pl_rect_h(rc)),
             block_w = sh->res.compute_group_size[0],
             block_h = sh->res.compute_group_size[1],
             num_x   = (width  + block_w - 1) / block_w,
@@ -901,12 +989,12 @@ bool pl_dispatch_finish(struct pl_dispatch *dp, struct pl_shader **psh,
         rparams->compute_groups[2] = 1;
     } else {
         // Update the scissors for performance
-        rparams->scissors = *rc;
-        pl_rect2d_normalize(&rparams->scissors);
+        rparams->scissors = rc_norm;
     }
 
     // Dispatch the actual shader
-    rparams->target = target;
+    rparams->target = params->target;
+    rparams->timer = params->timer;
     pl_pass_run(dp->gpu, &pass->run_params);
     ret = true;
 
@@ -915,14 +1003,14 @@ error:
     for (int i = 0; i < PL_ARRAY_SIZE(dp->tmp); i++)
         dp->tmp[i].len = 0;
 
-    pl_dispatch_abort(dp, psh);
+    pl_dispatch_abort(dp, params->shader);
     return ret;
 }
 
-bool pl_dispatch_compute(struct pl_dispatch *dp, struct pl_shader **psh,
-                         int dispatch_size[3])
+bool pl_dispatch_compute(struct pl_dispatch *dp,
+                         const struct pl_dispatch_compute_params *params)
 {
-    struct pl_shader *sh = *psh;
+    struct pl_shader *sh = *params->shader;
     const struct pl_shader_res *res = &sh->res;
     bool ret = false;
 
@@ -948,12 +1036,18 @@ bool pl_dispatch_compute(struct pl_dispatch *dp, struct pl_shader **psh,
     }
 
     if (sh->res.num_vertex_attribs) {
-        PL_ERR(dp, "Trying to dispatch a targetless compute shader that uses "
-               "vertex attributes!");
-        goto error;
+        if (!params->width || !params->height) {
+            PL_ERR(dp, "Trying to dispatch a targetless compute shader that "
+                   "uses vertex attributes, this requires specifying the size "
+                   "of the effective rendering area!");
+            goto error;
+        }
+
+        compute_vertex_attribs(dp, sh, params->width, params->height,
+                               &(ident_t){0});
     }
 
-    struct pass *pass = find_pass(dp, sh, NULL, NULL, NULL);
+    struct pass *pass = find_pass(dp, sh, NULL, NULL, NULL, false);
 
     // Silently return on failed passes
     if (pass->failed)
@@ -963,18 +1057,21 @@ bool pl_dispatch_compute(struct pl_dispatch *dp, struct pl_shader **psh,
 
     // Update the descriptor bindings
     for (int i = 0; i < sh->res.num_descriptors; i++)
-        rparams->desc_bindings[i].object = sh->res.descriptors[i].object;
+        rparams->desc_bindings[i].object = sh->descriptors[i].object;
 
     // Update all of the variables (if needed)
     rparams->num_var_updates = 0;
     for (int i = 0; i < res->num_variables; i++)
-        update_pass_var(dp, pass, &res->variables[i], &pass->vars[i]);
+        update_pass_var(dp, pass, &sh->variables[i], &pass->vars[i]);
 
     // Update the dispatch size
-    for (int i = 0; i < 3; i++)
-        rparams->compute_groups[i] = dispatch_size[i];
+    for (int i = 0; i < 3; i++) {
+        pl_assert(params->dispatch_size[i] > 0);
+        rparams->compute_groups[i] = params->dispatch_size[i];
+    }
 
     // Dispatch the actual shader
+    rparams->timer = params->timer;
     pl_pass_run(dp->gpu, &pass->run_params);
     ret = true;
 
@@ -983,7 +1080,7 @@ error:
     for (int i = 0; i < PL_ARRAY_SIZE(dp->tmp); i++)
         dp->tmp[i].len = 0;
 
-    pl_dispatch_abort(dp, psh);
+    pl_dispatch_abort(dp, params->shader);
     return ret;
 }
 
