@@ -29,7 +29,9 @@ struct pl_gl {
     int gl_ver;
     int gles_ver;
     bool has_stride;
+    bool has_invalidate;
     bool has_vao;
+    bool has_queries;
 };
 
 static bool test_ext(const struct pl_gpu *gpu, const char *ext,
@@ -233,7 +235,7 @@ const struct pl_gpu *pl_gpu_create_gl(struct pl_context *ctx)
     // Query all device limits
     struct pl_gpu_limits *l = &gpu->limits;
     get(GL_MAX_TEXTURE_SIZE, &l->max_tex_2d_dim);
-    if (test_ext(gpu, NULL, 21, 30)) // FIXME: is there an ext for this?
+    if (test_ext(gpu, "GL_EXT_texture3D", 21, 30))
         get(GL_MAX_3D_TEXTURE_SIZE, &l->max_tex_3d_dim);
 
     // There's no equivalent limit for 1D textures for whatever reason, so
@@ -241,18 +243,14 @@ const struct pl_gpu *pl_gpu_create_gl(struct pl_context *ctx)
     if (p->gl_ver >= 21)
         l->max_tex_1d_dim = l->max_tex_2d_dim;
 
-    if (test_ext(gpu, "GL_ARB_pixel_buffer_object", 0, 0)) // FIXME: when is this core?
+    if (test_ext(gpu, "GL_ARB_pixel_buffer_object", 31, 0))
         l->max_xfer_size = SIZE_MAX; // no limit imposed by GL
     if (test_ext(gpu, "GL_ARB_uniform_buffer_object", 31, 0))
         get(GL_MAX_UNIFORM_BLOCK_SIZE, &l->max_ubo_size);
-    if (test_ext(gpu, "GL_ARB_shader_storage_buffer_object", 43, 0) &&
-        test_ext(gpu, "GL_ARB_shader_image_load_store", 42, 0))
-    {
-        // storage image support is needed for glMemoryBarrier!
+    if (test_ext(gpu, "GL_ARB_shader_storage_buffer_object", 43, 0))
         get(GL_MAX_SHADER_STORAGE_BLOCK_SIZE, &l->max_ssbo_size);
-    }
 
-    if (test_ext(gpu, "GL_ARB_texture_gather", 0, 0)) { // FIXME: when is this core?
+    if (test_ext(gpu, "GL_ARB_texture_gather", 40, 0)) {
         get(GL_MIN_PROGRAM_TEXTURE_GATHER_OFFSET_ARB, &l->min_gather_offset);
         get(GL_MAX_PROGRAM_TEXTURE_GATHER_OFFSET_ARB, &l->max_gather_offset);
     }
@@ -269,6 +267,8 @@ const struct pl_gpu *pl_gpu_create_gl(struct pl_context *ctx)
     // Cached some existing capability checks
     p->has_stride = test_ext(gpu, "GL_EXT_unpack_subimage", 11, 30);
     p->has_vao = test_ext(gpu, "GL_ARB_vertex_array_object", 30, 0);
+    p->has_invalidate = test_ext(gpu, "GL_ARB_invalidate_subdata", 43, 30);
+    p->has_queries = test_ext(gpu, "GL_ARB_timer_query", 33, 0);
 
     // We simply don't know, so make up some values
     gpu->limits.align_tex_xfer_offset = 32;
@@ -295,9 +295,11 @@ error:
 struct pl_tex_gl {
     GLenum target;
     GLuint texture;
+    bool wrapped;
     GLint filter;
     GLuint fbo; // or 0
     bool wrapped_fb;
+    GLbitfield barrier;
 
     // GL format fields
     GLenum format;
@@ -308,14 +310,32 @@ struct pl_tex_gl {
 static void gl_tex_destroy(const struct pl_gpu *gpu, const struct pl_tex *tex)
 {
     struct pl_tex_gl *tex_gl = TA_PRIV(tex);
-    if (tex_gl->fbo)
+    if (tex_gl->fbo && !tex_gl->wrapped_fb)
         glDeleteFramebuffers(1, &tex_gl->fbo);
-    glDeleteTextures(1, &tex_gl->texture);
+    if (!tex_gl->wrapped)
+        glDeleteTextures(1, &tex_gl->texture);
 
     talloc_free((void *) tex);
     gl_check_err(gpu, "gl_tex_destroy");
 }
 
+static GLbitfield tex_barrier(const struct pl_tex *tex)
+{
+    GLbitfield barrier = 0;
+    const struct pl_tex_params *params = &tex->params;
+
+    if (params->sampleable)
+        barrier |= GL_TEXTURE_FETCH_BARRIER_BIT;
+    if (params->renderable || params->blit_src || params->blit_dst)
+        barrier |= GL_FRAMEBUFFER_BARRIER_BIT;
+    if (params->storable)
+        barrier |= GL_SHADER_IMAGE_ACCESS_BARRIER_BIT;
+    if (params->host_writable || params->host_readable)
+        barrier |= GL_TEXTURE_UPDATE_BARRIER_BIT;
+
+    return barrier;
+}
+
 static const struct pl_tex *gl_tex_create(const struct pl_gpu *gpu,
                                           const struct pl_tex_params *params)
 {
@@ -324,6 +344,7 @@ static const struct pl_tex *gl_tex_create(const struct pl_gpu *gpu,
     struct pl_tex *tex = talloc_zero_priv(NULL, struct pl_tex, struct pl_tex_gl);
     tex->params = *params;
     tex->params.initial_data = NULL;
+    tex->sampler_type = PL_SAMPLER_NORMAL;
 
     struct pl_tex_gl *tex_gl = TA_PRIV(tex);
 
@@ -332,6 +353,7 @@ static const struct pl_tex *gl_tex_create(const struct pl_gpu *gpu,
     tex_gl->format = fmt->fmt;
     tex_gl->iformat = fmt->ifmt;
     tex_gl->type = fmt->type;
+    tex_gl->barrier = tex_barrier(tex);
 
     static const GLint targets[] = {
         [1] = GL_TEXTURE_1D,
@@ -472,12 +494,15 @@ error:
 
 static bool gl_fb_query(const struct pl_gpu *gpu, int fbo, struct pl_fmt *fmt)
 {
+    struct pl_gl *p = TA_PRIV(gpu);
     if (!test_ext(gpu, "GL_ARB_framebuffer_object", 30, 20))
-        return true;
+        goto fallback;
+    if (!fbo && p->gles_ver && p->gles_ver < 30)
+        goto fallback; // can't query default framebuffer on GLES 2.0
 
     glBindFramebuffer(GL_FRAMEBUFFER, fbo);
 
-    GLenum obj = gpu->glsl.gles ? GL_BACK : GL_BACK_LEFT;
+    GLenum obj = p->gles_ver ? GL_BACK : GL_BACK_LEFT;
     if (fbo != 0)
         obj = GL_COLOR_ATTACHMENT0;
 
@@ -509,7 +534,17 @@ static bool gl_fb_query(const struct pl_gpu *gpu, int fbo, struct pl_fmt *fmt)
         bits += fmt->component_depth[i];
     fmt->internal_size = (bits + 7) / 8;
 
-    return gl_check_err(gpu, "gl_fb_query");
+    if (!gl_check_err(gpu, "gl_fb_query"))
+        goto fallback; // this happens on some outdated drivers
+
+    return true;
+
+fallback:
+    // default to rgba8
+    for (int i = 0; i < 4; i++)
+        fmt->component_depth[i] = 8;
+    fmt->internal_size = 4;
+    return true;
 }
 
 static const struct gl_format fbo_dummy_format = {
@@ -556,6 +591,7 @@ const struct pl_tex *pl_opengl_wrap_fb(const struct pl_gpu *gpu, GLuint fbo,
         .iformat = 0,
         .type = 0,
         .wrapped_fb = true,
+        .barrier = tex_barrier(tex),
     };
 
     return tex;
@@ -565,16 +601,207 @@ error:
     return NULL;
 }
 
+const struct pl_tex *pl_opengl_wrap(const struct pl_gpu *gpu,
+                                    const struct pl_opengl_wrap_params *params)
+{
+    struct pl_gl *p = TA_PRIV(gpu);
+
+    const struct pl_fmt *fmt = NULL;
+    const struct gl_format *glfmt = NULL;
+    for (int i = 0; i < gpu->num_formats; i++) {
+        const struct gl_format **glfmtp = TA_PRIV(gpu->formats[i]);
+        if ((*glfmtp)->ifmt == params->iformat) {
+            fmt = gpu->formats[i];
+            glfmt = *glfmtp;
+            break;
+        }
+    }
+
+    if (!fmt) {
+        PL_ERR(gpu, "Failed mapping iformat %d to any equivalent `pl_fmt`",
+               params->iformat);
+        return NULL;
+    }
+
+    enum pl_sampler_type sampler_type;
+    switch (params->target) {
+    case GL_TEXTURE_1D:
+        if (params->width || params->depth) {
+            PL_ERR(gpu, "Invalid texture dimensions for GL_TEXTURE_1D");
+            return NULL;
+        }
+        // fall through
+    case GL_TEXTURE_2D:
+        if (params->depth) {
+            PL_ERR(gpu, "Invalid texture dimensions for GL_TEXTURE_2D");
+            return NULL;
+        }
+        // fall through
+    case 0:
+    case GL_TEXTURE_3D:
+        sampler_type = PL_SAMPLER_NORMAL;
+        break;
+
+    case GL_TEXTURE_RECTANGLE: sampler_type = PL_SAMPLER_RECT; break;
+    case GL_TEXTURE_EXTERNAL_OES: sampler_type = PL_SAMPLER_EXTERNAL; break;
+
+    default:
+        PL_ERR(gpu, "Failed mapping texture target %u to any equivalent "
+               "`pl_sampler_type`", params->target);
+        return NULL;
+    }
+
+    enum pl_tex_sample_mode sample_mode;
+    switch (params->filter) {
+    case 0: // fall through
+    case GL_LINEAR: sample_mode = PL_TEX_SAMPLE_LINEAR; break;
+    case GL_NEAREST: sample_mode = PL_TEX_SAMPLE_NEAREST; break;
+
+    default:
+        PL_ERR(gpu, "Failed mapping texture filter %d to `pl_tex_sample_mode`",
+               params->filter);
+        return NULL;
+    }
+
+    enum pl_tex_address_mode address_mode;
+    switch (params->address_mode) {
+    case 0: // fall through
+    case GL_REPEAT: address_mode = PL_TEX_ADDRESS_REPEAT; break;
+    case GL_CLAMP_TO_EDGE: address_mode = PL_TEX_ADDRESS_CLAMP; break;
+    case GL_MIRRORED_REPEAT: address_mode = PL_TEX_ADDRESS_MIRROR; break;
+
+    default:
+        PL_ERR(gpu, "Failed mapping address mode %d to `pl_tex_address_mode`",
+               params->address_mode);
+        return NULL;
+    }
+
+    struct pl_tex *tex = talloc_priv(NULL, struct pl_tex, struct pl_tex_gl);
+    struct pl_tex_gl *tex_gl = TA_PRIV(tex);
+
+    *tex = (struct pl_tex) {
+        .sampler_type = sampler_type,
+        .params = {
+            .w = params->width,
+            .h = params->height,
+            .d = params->depth,
+            .format = fmt,
+            .sampleable = true,
+            .storable = fmt->caps & PL_FMT_CAP_STORABLE,
+            .host_writable = true,
+            .sample_mode = sample_mode,
+            .address_mode = address_mode,
+        },
+    };
+
+    *tex_gl = (struct pl_tex_gl) {
+        .target = params->target,
+        .texture = params->texture,
+        .wrapped = true,
+        .filter = PL_DEF(params->filter, GL_LINEAR),
+        .iformat = glfmt->ifmt,
+        .format = glfmt->fmt,
+        .type = glfmt->type,
+        .barrier = tex_barrier(tex),
+    };
+
+    int dims = pl_tex_params_dimension(tex->params);
+    if (!tex_gl->target) {
+        switch (dims) {
+        case 1: tex_gl->target = GL_TEXTURE_1D; break;
+        case 2: tex_gl->target = GL_TEXTURE_2D; break;
+        case 3: tex_gl->target = GL_TEXTURE_3D; break;
+        }
+    }
+
+    bool can_fbo = (fmt->caps & PL_FMT_CAP_RENDERABLE) &&
+                   sampler_type != PL_SAMPLER_EXTERNAL &&
+                   dims < 3;
+
+    if (can_fbo) {
+        glGenFramebuffers(1, &tex_gl->fbo);
+        glBindFramebuffer(GL_FRAMEBUFFER, tex_gl->fbo);
+        switch (dims) {
+        case 1:
+            glFramebufferTexture1D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
+                                   tex_gl->target, tex_gl->texture, 0);
+            break;
+        case 2:
+            glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
+                                   tex_gl->target, tex_gl->texture, 0);
+            break;
+        case 3: abort();
+        }
+
+        GLenum err = glCheckFramebufferStatus(GL_FRAMEBUFFER);
+        if (err != GL_FRAMEBUFFER_COMPLETE) {
+            glBindFramebuffer(GL_FRAMEBUFFER, 0);
+            PL_ERR(gpu, "Failed creating framebuffer: error code %d", err);
+            goto error;
+        }
+
+        tex->params.renderable = true;
+        if (dims == 2 && (fmt->caps & PL_FMT_CAP_BLITTABLE)) {
+            tex->params.blit_src = true;
+            tex->params.blit_dst = true;
+        }
+
+        if (p->gles_ver) {
+            GLint read_type = 0, read_fmt = 0;
+            glGetIntegerv(GL_IMPLEMENTATION_COLOR_READ_TYPE, &read_type);
+            glGetIntegerv(GL_IMPLEMENTATION_COLOR_READ_FORMAT, &read_fmt);
+            tex->params.host_readable = read_type == tex_gl->type &&
+                                        read_fmt == tex_gl->format;
+        } else {
+            tex->params.host_readable = true;
+        }
+
+        glBindFramebuffer(GL_FRAMEBUFFER, 0);
+        if (!gl_check_err(gpu, "pl_opengl_wrap: fbo"))
+            goto error;
+    }
+
+    return tex;
+
+error:
+    gl_tex_destroy(gpu, tex);
+    return NULL;
+}
+
+unsigned int pl_opengl_unwrap(const struct pl_gpu *gpu, const struct pl_tex *tex,
+                              unsigned int *out_target, int *out_iformat,
+                              unsigned int *out_fbo)
+{
+    struct pl_tex_gl *tex_gl = TA_PRIV(tex);
+    if (!tex_gl->texture) {
+        PL_ERR(gpu, "Trying to call `pl_opengl_unwrap` on a pseudo-texture "
+               "(perhaps obtained by `pl_swapchain_start_frame`?)");
+        return 0;
+    }
+
+    if (out_target)
+        *out_target = tex_gl->target;
+    if (out_iformat)
+        *out_iformat = tex_gl->iformat;
+    if (out_fbo)
+        *out_fbo = tex_gl->fbo;
+
+    return tex_gl->texture;
+}
+
 static void gl_tex_invalidate(const struct pl_gpu *gpu, const struct pl_tex *tex)
 {
     struct pl_gl *p = TA_PRIV(gpu);
     struct pl_tex_gl *tex_gl = TA_PRIV(tex);
 
-    if (tex_gl->wrapped_fb && (p->gl_ver >= 43 || p->gles_ver >= 30)) {
+    if (!p->has_invalidate)
+        return;
+
+    if (tex_gl->wrapped_fb) {
         glBindFramebuffer(GL_FRAMEBUFFER, tex_gl->fbo);
         glInvalidateFramebuffer(GL_FRAMEBUFFER, 1, (GLenum[]){GL_COLOR});
         glBindFramebuffer(GL_FRAMEBUFFER, 0);
-    } else if (p->gl_ver >= 43) {
+    } else {
         glInvalidateTexImage(tex_gl->texture, 0);
     }
 
@@ -620,6 +847,7 @@ struct pl_buf_gl {
     GLenum target;
     GLuint buffer;
     GLsync fence;
+    GLbitfield barrier;
 };
 
 static void gl_buf_destroy(const struct pl_gpu *gpu, const struct pl_buf *buf)
@@ -709,6 +937,14 @@ static const struct pl_buf *gl_buf_create(const struct pl_gpu *gpu,
     if (!gl_check_err(gpu, "gl_buf_create"))
         goto error;
 
+    if (params->type == PL_BUF_STORAGE) {
+        buf_gl->barrier = GL_SHADER_STORAGE_BARRIER_BIT;
+        if (params->host_writable || params->host_readable)
+            buf_gl->barrier |= GL_BUFFER_UPDATE_BARRIER_BIT;
+        if (params->host_mapped)
+            buf_gl->barrier |= GL_CLIENT_MAPPED_BUFFER_BARRIER_BIT;
+    }
+
     return buf;
 
 error:
@@ -778,6 +1014,9 @@ static int get_alignment(int stride)
     return 1;
 }
 
+static void gl_timer_begin(struct pl_timer *timer);
+static void gl_timer_end(struct pl_timer *timer);
+
 static bool gl_tex_upload(const struct pl_gpu *gpu,
                           const struct pl_tex_transfer_params *params)
 {
@@ -816,6 +1055,8 @@ static bool gl_tex_upload(const struct pl_gpu *gpu,
     }
 
     glBindTexture(tex_gl->target, tex_gl->texture);
+    gl_timer_begin(params->timer);
+
     switch (dims) {
     case 1:
         glTexSubImage1D(tex_gl->target, 0, params->rc.x0, pl_rect_w(params->rc),
@@ -839,6 +1080,7 @@ static bool gl_tex_upload(const struct pl_gpu *gpu,
         break;
     }
 
+    gl_timer_end(params->timer);
     glBindTexture(tex_gl->target, 0);
     glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
     if (p->has_stride) {
@@ -888,6 +1130,8 @@ static bool gl_tex_download(const struct pl_gpu *gpu,
                    params->stride_w == tex->params.w &&
                    params->stride_h == PL_DEF(tex->params.h, 1);
 
+    gl_timer_begin(params->timer);
+
     if (tex_gl->fbo || tex_gl->wrapped_fb) {
         // We can use a more efficient path when we have an FBO available
         if (dims > 1) {
@@ -926,6 +1170,8 @@ static bool gl_tex_download(const struct pl_gpu *gpu,
         ok = false;
     }
 
+    gl_timer_end(params->timer);
+
     if (buf) {
         glBindBuffer(GL_PIXEL_PACK_BUFFER, 0);
         if (ok && buf->params.host_mapped) {
@@ -1332,8 +1578,6 @@ static void update_desc(const struct pl_pass *pass, int index,
         const struct pl_buf *buf = db->object;
         struct pl_buf_gl *buf_gl = TA_PRIV(buf);
         glBindBufferBase(buf_gl->target, desc->binding, buf_gl->buffer);
-        // SSBOs are not implicitly coherent in OpenGL
-        glMemoryBarrier(buf_gl->target);
         break;
     }
     case PL_DESC_BUF_TEXEL_UNIFORM:
@@ -1356,15 +1600,22 @@ static void unbind_desc(const struct pl_pass *pass, int index,
         glBindTexture(tex_gl->target, 0);
         break;
     }
-    case PL_DESC_STORAGE_IMG:
+    case PL_DESC_STORAGE_IMG: {
+        const struct pl_tex *tex = db->object;
+        struct pl_tex_gl *tex_gl = TA_PRIV(tex);
         glBindImageTexture(desc->binding, 0, 0, GL_FALSE, 0,
                            GL_WRITE_ONLY, GL_R32F);
+        if (desc->access != PL_DESC_ACCESS_READONLY)
+            glMemoryBarrier(tex_gl->barrier);
         break;
+    }
     case PL_DESC_BUF_UNIFORM:
     case PL_DESC_BUF_STORAGE: {
         const struct pl_buf *buf = db->object;
         struct pl_buf_gl *buf_gl = TA_PRIV(buf);
         glBindBufferBase(buf_gl->target, desc->binding, 0);
+        if (desc->type == PL_DESC_BUF_STORAGE && desc->access != PL_DESC_ACCESS_READONLY)
+            glMemoryBarrier(buf_gl->barrier);
         break;
     }
     case PL_DESC_BUF_TEXEL_UNIFORM:
@@ -1379,6 +1630,7 @@ static void gl_pass_run(const struct pl_gpu *gpu,
 {
     const struct pl_pass *pass = params->pass;
     struct pl_pass_gl *pass_gl = TA_PRIV(pass);
+    struct pl_gl *p = TA_PRIV(gpu);
 
     glUseProgram(pass_gl->program);
 
@@ -1395,7 +1647,7 @@ static void gl_pass_run(const struct pl_gpu *gpu,
     case PL_PASS_RASTER: {
         struct pl_tex_gl *target_gl = TA_PRIV(params->target);
         glBindFramebuffer(GL_FRAMEBUFFER, target_gl->fbo);
-        if (!pass->params.load_target) {
+        if (!pass->params.load_target && p->has_invalidate) {
             GLenum fb = target_gl->fbo ? GL_COLOR_ATTACHMENT0 : GL_COLOR;
             glInvalidateFramebuffer(GL_FRAMEBUFFER, 1, &fb);
         }
@@ -1443,7 +1695,9 @@ static void gl_pass_run(const struct pl_gpu *gpu,
             [PL_PRIM_TRIANGLE_FAN]      = GL_TRIANGLE_FAN,
         };
 
+        gl_timer_begin(params->timer);
         glDrawArrays(map_prim[pass->params.vertex_type], 0, params->vertex_count);
+        gl_timer_end(params->timer);
         gl_check_err(gpu, "gl_pass_run: drawing");
 
         if (pass_gl->vao) {
@@ -1461,11 +1715,11 @@ static void gl_pass_run(const struct pl_gpu *gpu,
     }
 
     case PL_PASS_COMPUTE:
+        gl_timer_begin(params->timer);
         glDispatchCompute(params->compute_groups[0],
                           params->compute_groups[1],
                           params->compute_groups[2]);
-
-        glMemoryBarrier(GL_TEXTURE_FETCH_BARRIER_BIT);
+        gl_timer_end(params->timer);
         break;
 
     default: abort();
@@ -1479,6 +1733,71 @@ static void gl_pass_run(const struct pl_gpu *gpu,
     gl_check_err(gpu, "gl_pass_run");
 }
 
+#define QUERY_OBJECT_NUM 8
+
+struct pl_timer {
+    GLuint query[QUERY_OBJECT_NUM];
+    int index_write; // next index to write to
+    int index_read; // next index to read from
+};
+
+static struct pl_timer *gl_timer_create(const struct pl_gpu *gpu)
+{
+    struct pl_gl *p = TA_PRIV(gpu);
+    if (!p->has_queries)
+        return NULL;
+
+    struct pl_timer *timer = talloc_zero(NULL, struct pl_timer);
+    glGenQueries(QUERY_OBJECT_NUM, timer->query);
+    return timer;
+}
+
+static void gl_timer_destroy(const struct pl_gpu *gpu, struct pl_timer *timer)
+{
+    glDeleteQueries(QUERY_OBJECT_NUM, timer->query);
+    talloc_free(timer);
+}
+
+static uint64_t gl_timer_query(const struct pl_gpu *gpu, struct pl_timer *timer)
+{
+    if (timer->index_read == timer->index_write)
+        return 0; // no more unprocessed results
+
+    GLuint query = timer->query[timer->index_read];
+    int avail = 0;
+    glGetQueryObjectiv(query, GL_QUERY_RESULT_AVAILABLE, &avail);
+    if (!avail)
+        return 0;
+
+    uint64_t res = 0;
+    glGetQueryObjectui64v(query, GL_QUERY_RESULT, &res);
+
+    timer->index_read = (timer->index_read + 1) % QUERY_OBJECT_NUM;
+    return res;
+}
+
+static void gl_timer_begin(struct pl_timer *timer)
+{
+    if (!timer)
+        return;
+
+    glBeginQuery(GL_TIME_ELAPSED, timer->query[timer->index_write]);
+}
+
+static void gl_timer_end(struct pl_timer *timer)
+{
+    if (!timer)
+        return;
+
+    glEndQuery(GL_TIME_ELAPSED);
+
+    timer->index_write = (timer->index_write + 1) % QUERY_OBJECT_NUM;
+    if (timer->index_write == timer->index_read) {
+        // forcibly drop the least recent result to make space
+        timer->index_read = (timer->index_read + 1) % QUERY_OBJECT_NUM;
+    }
+}
+
 static void gl_gpu_flush(const struct pl_gpu *gpu)
 {
     glFlush();
@@ -1509,6 +1828,9 @@ static const struct pl_gpu_fns pl_fns_gl = {
     .pass_create            = gl_pass_create,
     .pass_destroy           = gl_pass_destroy,
     .pass_run               = gl_pass_run,
+    .timer_create           = gl_timer_create,
+    .timer_destroy          = gl_timer_destroy,
+    .timer_query            = gl_timer_query,
     .gpu_flush              = gl_gpu_flush,
     .gpu_finish             = gl_gpu_finish,
 };
