@@ -51,7 +51,7 @@ void pl_shader_decode_color(struct pl_shader *sh, struct pl_color_repr *repr,
     });
 
     ident_t cmat_c = sh_var(sh, (struct pl_shader_var) {
-        .var  = pl_var_vec3("cmat_m"),
+        .var  = pl_var_vec3("cmat_c"),
         .data = tr.c,
     });
 
@@ -151,16 +151,6 @@ void pl_shader_decode_color(struct pl_shader *sh, struct pl_color_repr *repr,
 void pl_shader_encode_color(struct pl_shader *sh,
                             const struct pl_color_repr *repr)
 {
-    // Since this is a relatively rare operation, bypass it as much as possible
-    bool skip = true;
-    skip &= PL_DEF(repr->sys, PL_COLOR_SYSTEM_RGB) == PL_COLOR_SYSTEM_RGB;
-    skip &= PL_DEF(repr->levels, PL_COLOR_LEVELS_PC) == PL_COLOR_LEVELS_PC;
-    skip &= PL_DEF(repr->alpha, PL_ALPHA_PREMULTIPLIED) == PL_ALPHA_PREMULTIPLIED;
-    skip &= repr->bits.sample_depth == repr->bits.color_depth;
-    skip &= !repr->bits.bit_shift;
-    if (skip)
-        return;
-
     if (!sh_require(sh, PL_SHADER_SIG_COLOR, 0, 0))
         return;
 
@@ -217,29 +207,41 @@ void pl_shader_encode_color(struct pl_shader *sh,
         break;
     }
 
-    struct pl_color_repr copy = *repr;
-    float xyzscale = (repr->sys == PL_COLOR_SYSTEM_XYZ)
-                        ? pl_color_repr_normalize(&copy)
-                        : 0.0;
+    // Since this is a relatively rare operation, bypass it as much as possible
+    bool skip = true;
+    skip &= PL_DEF(repr->sys, PL_COLOR_SYSTEM_RGB) == PL_COLOR_SYSTEM_RGB;
+    skip &= PL_DEF(repr->levels, PL_COLOR_LEVELS_PC) == PL_COLOR_LEVELS_PC;
+    skip &= PL_DEF(repr->bits.sample_depth, 8) == PL_DEF(repr->bits.color_depth, 8);
+    skip &= !repr->bits.bit_shift;
 
-    struct pl_transform3x3 tr = pl_color_repr_decode(&copy, NULL);
-    pl_transform3x3_invert(&tr);
+    if (!skip) {
+        struct pl_color_repr copy = *repr;
+        float xyzscale = (repr->sys == PL_COLOR_SYSTEM_XYZ)
+                            ? pl_color_repr_normalize(&copy)
+                            : 0.0;
 
-    ident_t cmat = sh_var(sh, (struct pl_shader_var) {
-        .var  = pl_var_mat3("cmat"),
-        .data = PL_TRANSPOSE_3X3(tr.mat.m),
-    });
+        struct pl_transform3x3 tr = pl_color_repr_decode(&copy, NULL);
+        pl_transform3x3_invert(&tr);
 
-    ident_t cmat_c = sh_var(sh, (struct pl_shader_var) {
-        .var  = pl_var_vec3("cmat_m"),
-        .data = tr.c,
-    });
+        ident_t cmat = sh_var(sh, (struct pl_shader_var) {
+            .var  = pl_var_mat3("cmat"),
+            .data = PL_TRANSPOSE_3X3(tr.mat.m),
+        });
+
+        ident_t cmat_c = sh_var(sh, (struct pl_shader_var) {
+            .var  = pl_var_vec3("cmat_c"),
+            .data = tr.c,
+        });
+
+        GLSL("color.rgb = %s * color.rgb + %s;\n", cmat, cmat_c);
+
+        if (repr->sys == PL_COLOR_SYSTEM_XYZ)
+            GLSL("color.rgb = pow(color.rgb, vec3(1.0/2.6)) * vec3(1.0/%f); \n", xyzscale);
+    }
 
-    GLSL("color.rgb = %s * color.rgb + %s;\n", cmat, cmat_c);
-    if (repr->sys == PL_COLOR_SYSTEM_XYZ)
-        GLSL("color.rgb = pow(color.rgb, vec3(1.0/2.6)) * vec3(1.0/%f); \n", xyzscale);
     if (repr->alpha == PL_ALPHA_INDEPENDENT)
         GLSL("color.rgb /= vec3(max(color.a, 1e-6));\n");
+
     GLSL("}\n");
 }
 
@@ -317,16 +319,18 @@ void pl_shader_linearize(struct pl_shader *sh, enum pl_color_transfer trc)
              "             / (vec3(%f) - vec3(%f) * color.rgb); \n"
              "color.rgb = pow(color.rgb, vec3(1.0/%f));         \n"
              // PQ's output range is 0-10000, but we need it to be relative to
-             // to PL_COLOR_REF_WHITE instead, so rescale
-             "color.rgb *= vec3(%f);\n",
-             PQ_M2, PQ_C1, PQ_C2, PQ_C3, PQ_M1, 10000 / PL_COLOR_REF_WHITE);
+             // to PL_COLOR_SDR_WHITE instead, so rescale
+             "color.rgb *= vec3(%f);                            \n",
+             PQ_M2, PQ_C1, PQ_C2, PQ_C3, PQ_M1, 10000.0 / PL_COLOR_SDR_WHITE);
         break;
     case PL_COLOR_TRC_HLG:
         GLSL("color.rgb = mix(vec3(4.0) * color.rgb * color.rgb,         \n"
              "                exp((color.rgb - vec3(%f)) * vec3(1.0/%f)) \n"
              "                    + vec3(%f),                            \n"
-             "                %s(lessThan(vec3(0.5), color.rgb)));       \n",
-             HLG_C, HLG_A, HLG_B, sh_bvec(sh, 3));
+             "                %s(lessThan(vec3(0.5), color.rgb)));       \n"
+             // Rescale from 0-12 to be relative to PL_COLOR_SDR_WHITE_HLG
+             "color.rgb *= vec3(1.0/%f);                                 \n",
+             HLG_C, HLG_A, HLG_B, sh_bvec(sh, 3), PL_COLOR_SDR_WHITE_HLG);
         break;
     case PL_COLOR_TRC_V_LOG:
         GLSL("color.rgb = mix((color.rgb - vec3(0.125)) * vec3(1.0/5.6), \n"
@@ -397,13 +401,14 @@ void pl_shader_delinearize(struct pl_shader *sh, enum pl_color_transfer trc)
              "color.rgb = (vec3(%f) + vec3(%f) * color.rgb)      \n"
              "             / (vec3(1.0) + vec3(%f) * color.rgb); \n"
              "color.rgb = pow(color.rgb, vec3(%f));              \n",
-             10000 / PL_COLOR_REF_WHITE, PQ_M1, PQ_C1, PQ_C2, PQ_C3, PQ_M2);
+             10000 / PL_COLOR_SDR_WHITE, PQ_M1, PQ_C1, PQ_C2, PQ_C3, PQ_M2);
         break;
     case PL_COLOR_TRC_HLG:
-        GLSL("color.rgb = mix(vec3(0.5) * sqrt(color.rgb),                     \n"
+        GLSL("color.rgb *= vec3(%f);                                           \n"
+             "color.rgb = mix(vec3(0.5) * sqrt(color.rgb),                     \n"
              "                vec3(%f) * log(color.rgb - vec3(%f)) + vec3(%f), \n"
              "                %s(lessThan(vec3(1.0), color.rgb)));             \n",
-             HLG_A, HLG_B, HLG_C, sh_bvec(sh, 3));
+             PL_COLOR_SDR_WHITE_HLG, HLG_A, HLG_B, HLG_C, sh_bvec(sh, 3));
         break;
     case PL_COLOR_TRC_V_LOG:
         GLSL("color.rgb = mix(vec3(5.6) * color.rgb + vec3(0.125),       \n"
@@ -502,10 +507,11 @@ static void pl_shader_ootf(struct pl_shader *sh, struct pl_color_space csp)
     {
     case PL_COLOR_LIGHT_SCENE_HLG: {
         // HLG OOTF from BT.2100, tuned to the indicated peak
-        float gamma = 1.2 + 0.42 * log10(csp.sig_peak * PL_COLOR_REF_WHITE / 1000.0);
+        float peak = csp.sig_peak * PL_COLOR_SDR_WHITE;
+        float gamma = 1.2 + 0.42 * log10(peak / 1000.0);
         gamma = PL_MAX(gamma, 1.0);
         GLSL("color.rgb *= vec3(%f * pow(dot(%s, color.rgb), %f));\n",
-             csp.sig_peak / pow(12, gamma),
+             csp.sig_peak / pow(12.0 / PL_COLOR_SDR_WHITE_HLG, gamma),
              sh_luma_coeffs(sh, csp.primaries),
              gamma - 1.0);
         break;
@@ -540,12 +546,13 @@ static void pl_shader_inverse_ootf(struct pl_shader *sh, struct pl_color_space c
     switch (csp.light)
     {
     case PL_COLOR_LIGHT_SCENE_HLG: {
-        float gamma = 1.2 + 0.42 * log10(csp.sig_peak * PL_COLOR_REF_WHITE / 1000.0);
+        float peak = csp.sig_peak * PL_COLOR_SDR_WHITE;
+        float gamma = 1.2 + 0.42 * log10(peak / 1000.0);
         gamma = PL_MAX(gamma, 1.0);
         GLSL("color.rgb *= vec3(1.0/%f);                                \n"
              "color.rgb /= vec3(max(1e-6, pow(dot(%s, color.rgb),       \n"
              "                                %f)));                    \n",
-             csp.sig_peak / pow(12, gamma),
+             csp.sig_peak / pow(12.0 / PL_COLOR_SDR_WHITE_HLG, gamma),
              sh_luma_coeffs(sh, csp.primaries),
              (gamma - 1.0) / gamma);
         break;
@@ -574,12 +581,14 @@ const struct pl_peak_detect_params pl_peak_detect_default_params = {
     .smoothing_period       = 100.0,
     .scene_threshold_low    = 5.5,
     .scene_threshold_high   = 10.0,
+    .overshoot_margin       = 0.05,
 };
 
 struct sh_peak_obj {
     const struct pl_gpu *gpu;
     const struct pl_buf *buf;
     struct pl_shader_desc desc;
+    float margin;
 };
 
 static void sh_peak_uninit(const struct pl_gpu *gpu, void *ptr)
@@ -600,8 +609,6 @@ bool pl_shader_detect_peak(struct pl_shader *sh,
                            struct pl_shader_obj **state,
                            const struct pl_peak_detect_params *params)
 {
-    const struct pl_gpu *gpu = SH_GPU(sh);
-
     params = PL_DEF(params, &pl_peak_detect_default_params);
     if (!sh_require(sh, PL_SHADER_SIG_COLOR, 0, 0))
         return false;
@@ -611,7 +618,7 @@ bool pl_shader_detect_peak(struct pl_shader *sh,
         return false;
     }
 
-    if (gpu->glsl.version < 130) {
+    if (sh_glsl(sh).version < 130) {
         // uint was added in GLSL 130
         PL_ERR(sh, "HDR peak detection requires GLSL >= 130!");
         return false;
@@ -623,7 +630,9 @@ bool pl_shader_detect_peak(struct pl_shader *sh,
     if (!obj)
         return false;
 
+    const struct pl_gpu *gpu = SH_GPU(sh);
     obj->gpu = gpu;
+    obj->margin = params->overshoot_margin;
 
     if (!obj->buf) {
         obj->desc = (struct pl_shader_desc) {
@@ -657,7 +666,7 @@ bool pl_shader_detect_peak(struct pl_shader *sh,
     }
 
     if (!obj->buf) {
-        PL_ERR(sh, "Failed creating peak detection SSBO!");
+        SH_FAIL(sh, "Failed creating peak detection SSBO!");
         return false;
     }
 
@@ -745,9 +754,18 @@ bool pl_shader_detect_peak(struct pl_shader *sh,
     return true;
 }
 
+static inline float pq_delinearize(float x)
+{
+    x *= PL_COLOR_SDR_WHITE / 10000.0;
+    x = powf(x, PQ_M1);
+    x = (PQ_C1 + PQ_C2 * x) / (1.0 + PQ_C3 * x);
+    x = pow(x, PQ_M2);
+    return x;
+}
+
 const struct pl_color_map_params pl_color_map_default_params = {
     .intent                 = PL_INTENT_RELATIVE_COLORIMETRIC,
-    .tone_mapping_algo      = PL_TONE_MAPPING_HABLE,
+    .tone_mapping_algo      = PL_TONE_MAPPING_BT_2390,
     .desaturation_strength  = 0.75,
     .desaturation_exponent  = 1.50,
     .desaturation_base      = 0.18,
@@ -782,14 +800,24 @@ static void pl_shader_tone_map(struct pl_shader *sh, struct pl_color_space src,
             sh_desc(sh, obj->desc);
             GLSL("sig_avg  = average.x; \n"
                  "sig_peak = average.y; \n");
+
+            // Apply a tiny bit of extra margin of error for overshoot to the
+            // smoothed peak values, clamped to the maximum reasonable range.
+            if (obj->margin > 0.0) {
+                GLSL("sig_peak = min(sig_peak * %f, %f); \n",
+                     1.0 + obj->margin,
+                     10000 / PL_COLOR_SDR_WHITE);
+            }
         }
     }
 
-    // Rescale the input in order to bring it into a representation where
-    // 1.0 represents the dst_peak. This is because all of the tone mapping
-    // algorithms are defined in such a way that they map to the range [0.0, 1.0].
+    // Rescale the input in order to bring it into a representation where 1.0
+    // represents the dst_peak. This is because (almost) all of the tone
+    // mapping algorithms are defined in such a way that they map to the range
+    // [0.0, 1.0].
+    bool need_norm = params->tone_mapping_algo != PL_TONE_MAPPING_BT_2390;
     float dst_range = dst.sig_peak * dst.sig_scale;
-    if (dst_range > 1.0) {
+    if (dst_range > 1.0 && need_norm) {
         GLSL("color.rgb *= 1.0 / %f; \n"
              "sig_peak *= 1.0 / %f;  \n",
              dst_range, dst_range);
@@ -867,6 +895,41 @@ static void pl_shader_tone_map(struct pl_shader *sh, struct pl_color_space src,
         GLSL("sig *= %f / sig_peak;\n", PL_DEF(param, 1.0));
         break;
 
+    case PL_TONE_MAPPING_BT_2390:
+        // We first need to encode both sig and sig_peak into PQ space
+        GLSL("vec4 sig_pq = vec4(sig.rgb, sig_peak);                            \n"
+             "sig_pq *= vec4(1.0/%f);                                           \n"
+             "sig_pq = pow(sig_pq, vec4(%f));                                   \n"
+             "sig_pq = (vec4(%f) + vec4(%f) * sig_pq)                           \n"
+             "          / (vec4(1.0) + vec4(%f) * sig_pq);                      \n"
+             "sig_pq = pow(sig_pq, vec4(%f));                                   \n",
+             10000 / PL_COLOR_SDR_WHITE, PQ_M1, PQ_C1, PQ_C2, PQ_C3, PQ_M2);
+        // Encode both the signal and the target brightness to be relative to
+        // the source peak brightness, and figure out the target peak in this space
+        GLSL("float scale = 1.0 / sig_pq.a;                                     \n"
+             "sig_pq.rgb *= vec3(scale);                                        \n"
+             "float maxLum = %f * scale;                                        \n",
+             pq_delinearize(dst_range));
+        // Apply piece-wise hermite spline
+        GLSL("float ks = 1.5 * maxLum - 0.5;                                    \n"
+             "vec3 tb = (sig_pq.rgb - vec3(ks)) / vec3(1.0 - ks);               \n"
+             "vec3 tb2 = tb * tb;                                               \n"
+             "vec3 tb3 = tb2 * tb;                                              \n"
+             "vec3 pb = (2.0 * tb3 - 3.0 * tb2 + vec3(1.0)) * vec3(ks) +        \n"
+             "          (tb3 - 2.0 * tb2 + tb) * vec3(1.0 - ks) +               \n"
+             "          (-2.0 * tb3 + 3.0 * tb2) * vec3(maxLum);                \n"
+             "sig = mix(pb, sig_pq.rgb, %s(lessThan(sig_pq.rgb, vec3(ks))));    \n",
+             sh_bvec(sh, 3));
+        // Convert back from PQ space to linear light
+        GLSL("sig *= vec3(sig_pq.a);                                            \n"
+             "sig = pow(sig, vec3(1.0/%f));                                     \n"
+             "sig = max(sig - vec3(%f), 0.0) /                                  \n"
+             "          (vec3(%f) - vec3(%f) * sig);                            \n"
+             "sig = pow(sig, vec3(1.0/%f));                                     \n"
+             "sig *= vec3(%f);                                                  \n",
+             PQ_M2, PQ_C1, PQ_C2, PQ_C3, PQ_M1, 10000.0 / PL_COLOR_SDR_WHITE);
+        break;
+
     default:
         abort();
     }
@@ -889,7 +952,7 @@ static void pl_shader_tone_map(struct pl_shader *sh, struct pl_color_space src,
     }
 
     // Undo the normalization by `dst_peak`
-    if (dst_range > 1.0)
+    if (dst_range > 1.0 && need_norm)
         GLSL("color.rgb *= %f; \n", dst_range);
 }
 
@@ -1115,7 +1178,7 @@ void pl_shader_dither(struct pl_shader *sh, int new_depth,
 
         lut_size = 1 << PL_DEF(params->lut_size, 6);
         lut = sh_lut(sh, &obj->lut, SH_LUT_AUTO, lut_size, lut_size, 0, 1,
-                     changed, obj, fill_dither_matrix);
+                     changed, false, obj, fill_dither_matrix);
         if (!lut)
             goto fallback;
     }
@@ -1200,7 +1263,7 @@ const struct pl_dither_params pl_dither_default_params = {
     .temporal   = false, // commonly flickers on LCDs
 };
 
-#if PL_HAVE_LCMS
+#ifdef PL_HAVE_LCMS
 
 #include "lcms.h"
 
@@ -1268,7 +1331,7 @@ bool pl_3dlut_update(struct pl_shader *sh,
     obj->src = *src;
     obj->dst = *dst;
     obj->lut = sh_lut(sh, &obj->lut_obj, SH_LUT_LINEAR, s_r, s_g, s_b, 4,
-                      changed, obj, fill_3dlut);
+                      changed, false, obj, fill_3dlut);
     if (!obj->lut || !obj->ok)
         return false;
 
@@ -1296,11 +1359,11 @@ void pl_3dlut_apply(struct pl_shader *sh, struct pl_shader_obj **lut3d)
     obj->updated = false;
 }
 
+#endif // PL_HAVE_LCMS
+
 const struct pl_3dlut_params pl_3dlut_default_params = {
     .intent = PL_INTENT_RELATIVE_COLORIMETRIC,
     .size_r = 64,
     .size_g = 64,
     .size_b = 64,
 };
-
-#endif // PL_HAVE_LCMS
