@@ -25,20 +25,45 @@ const struct pl_deband_params pl_deband_default_params = {
     .grain      = 6.0,
 };
 
+static inline struct pl_tex_params src_params(const struct pl_sample_src *src)
+{
+    if (src->tex)
+        return src->tex->params;
+
+    return (struct pl_tex_params) {
+        .w = src->tex_w,
+        .h = src->tex_h,
+        .sample_mode = src->mode,
+    };
+}
+
 // Helper function to compute the src/dst sizes and upscaling ratios
 static bool setup_src(struct pl_shader *sh, const struct pl_sample_src *src,
                       ident_t *src_tex, ident_t *pos, ident_t *size, ident_t *pt,
                       float *ratio_x, float *ratio_y, int *components,
                       float *scale, bool resizeable, const char **fn)
 {
-    pl_assert(pl_tex_params_dimension(src->tex->params) == 2);
-    float src_w = pl_rect_w(src->rect);
-    float src_h = pl_rect_h(src->rect);
-    src_w = PL_DEF(src_w, src->tex->params.w);
-    src_h = PL_DEF(src_h, src->tex->params.h);
+    enum pl_shader_sig sig;
+    float src_w, src_h;
+    if (src->tex) {
+        pl_assert(pl_tex_params_dimension(src->tex->params) == 2);
+        sig = PL_SHADER_SIG_NONE;
+        src_w = pl_rect_w(src->rect);
+        src_h = pl_rect_h(src->rect);
+    } else {
+        pl_assert(src->tex_w && src->tex_h);
+        sig = PL_SHADER_SIG_SAMPLER;
+        src_w = src->sampled_w;
+        src_h = src->sampled_h;
+    }
 
-    int out_w = PL_DEF(src->new_w, fabs(src_w));
-    int out_h = PL_DEF(src->new_h, fabs(src_h));
+    src_w = PL_DEF(src_w, src_params(src).w);
+    src_h = PL_DEF(src_h, src_params(src).h);
+    pl_assert(src_w && src_h);
+
+    int out_w = PL_DEF(src->new_w, roundf(fabs(src_w)));
+    int out_h = PL_DEF(src->new_h, roundf(fabs(src_h)));
+    pl_assert(out_w && out_h);
 
     if (ratio_x)
         *ratio_x = out_w / fabs(src_w);
@@ -48,33 +73,74 @@ static bool setup_src(struct pl_shader *sh, const struct pl_sample_src *src,
         *scale = PL_DEF(src->scale, 1.0);
 
     if (components) {
-        const struct pl_fmt *fmt = src->tex->params.format;
-        *components = PL_DEF(src->components, fmt->num_components);
+        int tex_comps = 4;
+        if (src->tex)
+            tex_comps = src->tex->params.format->num_components;
+        *components = PL_DEF(src->components, tex_comps);
     }
 
     if (resizeable)
         out_w = out_h = 0;
-    if (!sh_require(sh, PL_SHADER_SIG_NONE, out_w, out_h))
+    if (!sh_require(sh, sig, out_w, out_h))
         return false;
 
-    struct pl_rect2df rect = {
-        .x0 = src->rect.x0,
-        .y0 = src->rect.y0,
-        .x1 = src->rect.x0 + src_w,
-        .y1 = src->rect.y0 + src_h,
-    };
+    if (src->tex) {
+        struct pl_rect2df rect = {
+            .x0 = src->rect.x0,
+            .y0 = src->rect.y0,
+            .x1 = src->rect.x0 + src_w,
+            .y1 = src->rect.y0 + src_h,
+        };
 
-    if (fn)
-        *fn = sh_tex_fn(sh, src->tex);
+        if (fn)
+            *fn = sh_tex_fn(sh, src->tex->params);
+
+        *src_tex = sh_bind(sh, src->tex, "src_tex", &rect, pos, size, pt);
+    } else {
+        if (size) {
+            *size = sh_var(sh, (struct pl_shader_var) {
+                .var = pl_var_vec2("tex_size"),
+                .data = &(float[2]) { src->tex_w, src->tex_h },
+            });
+        }
+
+        if (pt) {
+            float sx = 1.0 / src->tex_w, sy = 1.0 / src->tex_h;
+            if (src->sampler == PL_SAMPLER_RECT)
+                sx = sy = 1.0;
+
+            *pt = sh_var(sh, (struct pl_shader_var) {
+                .var = pl_var_vec2("tex_pt"),
+                .data = &(float[2]) { sx, sy },
+            });
+        }
+
+        if (fn)
+            *fn = sh_tex_fn(sh, (struct pl_tex_params) { .w = 1, .d = 1 }); // 2D
+
+        sh->sampler_type = src->sampler;
+
+        pl_assert(src->format);
+        switch (src->format) {
+        case PL_FMT_FLOAT:
+        case PL_FMT_UNORM:
+        case PL_FMT_SNORM: sh->sampler_prefix = ' '; break;
+        case PL_FMT_UINT: sh->sampler_prefix = 'u'; break;
+        case PL_FMT_SINT: sh->sampler_prefix = 's'; break;
+        default: abort();
+        }
+
+        *src_tex = "src_tex";
+        *pos = "tex_coord";
+    }
 
-    *src_tex = sh_bind(sh, src->tex, "src_tex", &rect, pos, size, pt);
     return true;
 }
 
 void pl_shader_deband(struct pl_shader *sh, const struct pl_sample_src *src,
                       const struct pl_deband_params *params)
 {
-    if (src->tex->params.sample_mode != PL_TEX_SAMPLE_LINEAR) {
+    if (src_params(src).sample_mode != PL_TEX_SAMPLE_LINEAR) {
         SH_FAIL(sh, "Debanding requires sample_mode = PL_TEX_SAMPLE_LINEAR!");
         return;
     }
@@ -174,7 +240,7 @@ static void bicubic_calcweights(struct pl_shader *sh, const char *t, const char
 
 bool pl_shader_sample_bicubic(struct pl_shader *sh, const struct pl_sample_src *src)
 {
-    if (src->tex->params.sample_mode != PL_TEX_SAMPLE_LINEAR) {
+    if (src_params(src).sample_mode != PL_TEX_SAMPLE_LINEAR) {
         SH_FAIL(sh, "Trying to use fast bicubic sampling from a texture without "
                 "PL_TEX_SAMPLE_LINEAR");
         return false;
@@ -313,16 +379,28 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
     }
 
     const struct pl_gpu *gpu = SH_GPU(sh);
-    const struct pl_tex *tex = src->tex;
-    pl_assert(gpu && tex);
+    pl_assert(gpu);
 
     bool has_compute = gpu->caps & PL_GPU_CAP_COMPUTE && !params->no_compute;
+    if (!src->tex && has_compute) {
+        // FIXME: Could maybe solve this by communicating the wbase from
+        // invocation 0 to the rest of the workgroup using shmem, which would
+        // also allow us to avoid the use of the hacky %s_map below.
+        PL_WARN(sh, "Combining pl_shader_sample_polar with the sampler2D "
+                "interface prevents the use of compute shaders, which is a "
+                "potentially massive performance hit. If you're sure you want "
+                "this, set `params.no_compute` to suppress this warning.");
+        has_compute = false;
+    }
+
     bool flipped = src->rect.x0 > src->rect.x1 || src->rect.y0 > src->rect.y1;
     if (flipped && has_compute) {
+        // FIXME: I'm sure this case could actually be supported with some
+        // extra math in the positional calculations, should implement it
         PL_WARN(sh, "Trying to use a flipped src.rect with polar sampling! "
                 "This prevents the use of compute shaders, which is a "
                 "potentially massive performance hit. If you're really sure you "
-                "want this, set params.no_compute to suppress this warning.");
+                "want this, set `params.no_compute` to suppress this warning.");
         has_compute = false;
     }
 
@@ -367,7 +445,7 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
     }
 
     ident_t lut = sh_lut(sh, &obj->lut, SH_LUT_LINEAR, lut_entries, 0, 0, 1,
-                         update, obj, fill_polar_lut);
+                         update, false, obj, fill_polar_lut);
     if (!lut) {
         SH_FAIL(sh, "Failed initializing polar LUT!");
         return false;
@@ -388,15 +466,10 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
     int offset  = bound - 1; // padding top/left
     int padding = offset + bound; // total padding
 
-    // For performance we want to load at least as many pixels horizontally as
-    // there are threads in a warp, as well as enough to take advantage of
-    // shmem parallelism. However, on the other hand, to hide latency we want
-    // to avoid making the kernel too large. A good size overall is 256
-    // threads, which allows at least 8 to run in parallel assuming good VGPR
-    // distribution. A good trade-off for the horizontal row size is 32, which
-    // is the warp size on nvidia. Going up to 64 (AMD's wavefront size)
-    // is not worth it even on AMD hardware.
-    const int bw = 32, bh = 256 / bw;
+    // Determined experimentally on modern AMD and Nvidia hardware. 32 is a
+    // good tradeoff for the horizontal work group size. Apart from that,
+    // just use as many threads as possible.
+    const int bw = 32, bh = gpu->limits.max_group_threads / bw;
 
     // We need to sample everything from base_min to base_max, so make sure
     // we have enough room in shmem
@@ -519,18 +592,17 @@ bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
     }
 
     const struct pl_gpu *gpu = SH_GPU(sh);
-    const struct pl_tex *tex = src->tex;
-    pl_assert(gpu && tex);
+    pl_assert(gpu);
 
     struct pl_sample_src srcfix = *src;
     switch (pass) {
     case PL_SEP_VERT:
         srcfix.rect.x0 = 0;
-        srcfix.rect.x1 = srcfix.new_w = tex->params.w;
+        srcfix.rect.x1 = srcfix.new_w = src_params(src).w;
         break;
     case PL_SEP_HORIZ:
         srcfix.rect.y0 = 0;
-        srcfix.rect.y1 = srcfix.new_h = tex->params.h;
+        srcfix.rect.y1 = srcfix.new_h = src_params(src).h;
         break;
     case PL_SEP_PASSES:
     default:
@@ -597,7 +669,7 @@ bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
     int N = obj->filter->row_size; // number of samples to convolve
     int width = obj->filter->row_stride / 4; // width of the LUT texture
     ident_t lut = sh_lut(sh, &obj->lut, SH_LUT_LINEAR, width, lut_entries, 0, 4,
-                         update, obj, fill_ortho_lut);
+                         update, false, obj, fill_ortho_lut);
     if (!lut) {
         SH_FAIL(sh, "Failed initializing separated LUT!");
         return false;
