@@ -37,6 +37,7 @@ struct priv {
     bool suboptimal;        // true once VK_SUBOPTIMAL_KHR is returned
     struct pl_color_repr color_repr;
     struct pl_color_space color_space;
+    struct pl_hdr_metadata hdr_metadata;
 
     // state of the images:
     const struct pl_tex **images; // pl_tex wrappers for the VkImages
@@ -137,13 +138,8 @@ static bool vk_map_color_space(VkColorSpaceKHR space, struct pl_color_space *out
         return false;
 #endif
 
-    // Included to satisfy the switch coverage check
-    case VK_COLOR_SPACE_RANGE_SIZE_KHR:
-    case VK_COLOR_SPACE_MAX_ENUM_KHR:
-        break;
+    default: return false;
     }
-
-    return false;
 }
 
 static bool pick_surf_format(const struct pl_gpu *gpu, const struct vk_ctx *vk,
@@ -164,7 +160,7 @@ static bool pick_surf_format(const struct pl_gpu *gpu, const struct vk_ctx *vk,
     }
 
     VK(vk->GetPhysicalDeviceSurfaceFormatsKHR(vk->physd, surf, &num, NULL));
-    formats = talloc_array(NULL, VkSurfaceFormatKHR, num);
+    formats = talloc_zero_array(NULL, VkSurfaceFormatKHR, num);
     VK(vk->GetPhysicalDeviceSurfaceFormatsKHR(vk->physd, surf, &num, formats));
 
     PL_DEBUG(gpu, "Available surface formats:");
@@ -210,9 +206,14 @@ static bool pick_surf_format(const struct pl_gpu *gpu, const struct vk_ctx *vk,
                 break; // accept
             continue;
 
-        // Accept 10/16 bit formats universally
+        // Only accept 10 bit formats for non-linear curves
         case VK_FORMAT_A2R10G10B10_UNORM_PACK32:
         case VK_FORMAT_A2B10G10R10_UNORM_PACK32:
+            if (space->transfer != PL_COLOR_TRC_LINEAR)
+                break; // accept
+            continue;
+
+        // Accept 16-bit formats for everything
         case VK_FORMAT_R16G16B16_UNORM:
         case VK_FORMAT_R16G16B16A16_UNORM:
              break; // accept
@@ -253,11 +254,19 @@ const struct pl_swapchain *pl_vulkan_create_swapchain(const struct pl_vulkan *pl
     struct vk_ctx *vk = TA_PRIV(plvk);
     const struct pl_gpu *gpu = plvk->gpu;
 
+    if (!vk->CreateSwapchainKHR) {
+        PL_ERR(gpu, VK_KHR_SWAPCHAIN_EXTENSION_NAME " not enabled!");
+        return NULL;
+    }
+
     VkSurfaceFormatKHR sfmt = params->surface_format;
     struct pl_color_space csp;
     if (!pick_surf_format(gpu, vk, params->surface, &sfmt, &csp))
         return NULL;
 
+    PL_DEBUG(gpu, "Picked surface format 0x%x, space 0x%x",
+             sfmt.format, sfmt.colorSpace);
+
     struct pl_swapchain *sw = talloc_zero_priv(NULL, struct pl_swapchain, struct priv);
     sw->impl = &vulkan_swapchain;
     sw->ctx = vk->ctx;
@@ -290,10 +299,10 @@ const struct pl_swapchain *pl_vulkan_create_swapchain(const struct pl_vulkan *pl
 
     // Make sure the swapchain present mode is supported
     VkPresentModeKHR *modes = NULL;
-    int num_modes;
+    int num_modes = 0;
     VK(vk->GetPhysicalDeviceSurfacePresentModesKHR(vk->physd, p->surf,
                                                    &num_modes, NULL));
-    modes = talloc_array(NULL, VkPresentModeKHR, num_modes);
+    modes = talloc_zero_array(NULL, VkPresentModeKHR, num_modes);
     VK(vk->GetPhysicalDeviceSurfacePresentModesKHR(vk->physd, p->surf,
                                                    &num_modes, modes));
 
@@ -347,7 +356,7 @@ static bool update_swapchain_info(struct priv *p, VkSwapchainCreateInfoKHR *info
     struct vk_ctx *vk = p->vk;
 
     // Query the supported capabilities and update this struct as needed
-    VkSurfaceCapabilitiesKHR caps;
+    VkSurfaceCapabilitiesKHR caps = {0};
     VK(vk->GetPhysicalDeviceSurfaceCapabilitiesKHR(vk->physd, p->surf, &caps));
 
     // Check for hidden/invisible window
@@ -370,6 +379,8 @@ static bool update_swapchain_info(struct priv *p, VkSwapchainCreateInfoKHR *info
         if (caps.supportedCompositeAlpha & alphaModes[i].vk_mode) {
             info->compositeAlpha = alphaModes[i].vk_mode;
             p->color_repr.alpha = alphaModes[i].pl_mode;
+            PL_DEBUG(vk, "Requested alpha compositing mode: 0x%x",
+                     info->compositeAlpha);
             break;
         }
     }
@@ -391,6 +402,7 @@ static bool update_swapchain_info(struct priv *p, VkSwapchainCreateInfoKHR *info
     for (int i = 0; i < PL_ARRAY_SIZE(rotModes); i++) {
         if (caps.supportedTransforms & rotModes[i]) {
             info->preTransform = rotModes[i];
+            PL_DEBUG(vk, "Requested surface transform: 0x%x", info->preTransform);
             break;
         }
     }
@@ -435,31 +447,33 @@ static bool update_swapchain_info(struct priv *p, VkSwapchainCreateInfoKHR *info
     h = PL_MIN(PL_MAX(h, caps.minImageExtent.height), caps.maxImageExtent.height);
     info->imageExtent = (VkExtent2D) { w, h };
 
-    // We just request whatever usage we can, and let the pl_vk decide what
-    // pl_tex_params that translates to. This makes the images as flexible
-    // as possible. However, require at least blitting and rendering.
-    VkImageUsageFlags required_flags = 0;
-    required_flags |= VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT;
-    required_flags |= VK_IMAGE_USAGE_TRANSFER_DST_BIT;
-    if ((caps.supportedUsageFlags & required_flags) != required_flags) {
-        PL_ERR(vk, "The swapchain doesn't support rendering and blitting!");
-        goto error;
-    }
-
-
-    // We need to intersect the swapchain supported usage flags with the
-    // format supported usage flags
+    // We just request whatever we can, and let the pl_vk decide what
+    // pl_tex_params that translates to. This makes the image as flexible as
+    // possible. That said, we still need to intersect the swapchain supported
+    // usage flags with the format supported usage flags
     info->imageUsage = caps.supportedUsageFlags;
-    VkFormatProperties fmtprop;
+    VkFormatProperties fmtprop = {0};
     vk->GetPhysicalDeviceFormatProperties(vk->physd, info->imageFormat, &fmtprop);
 
 #define CHECK(usage, feature) \
     if (!((fmtprop.optimalTilingFeatures & VK_FORMAT_FEATURE_##feature##_BIT))) \
         info->imageUsage &= ~VK_IMAGE_USAGE_##usage##_BIT
 
+    CHECK(TRANSFER_SRC, TRANSFER_SRC);
+    CHECK(TRANSFER_DST, TRANSFER_DST);
     CHECK(SAMPLED, SAMPLED_IMAGE);
-    CHECK(COLOR_ATTACHMENT, COLOR_ATTACHMENT);
     CHECK(STORAGE, STORAGE_IMAGE);
+    CHECK(COLOR_ATTACHMENT, COLOR_ATTACHMENT);
+
+    // Make sure we support at least rendering and blitting!
+    VkImageUsageFlags required_flags = 0;
+    required_flags |= VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT;
+    required_flags |= VK_IMAGE_USAGE_TRANSFER_DST_BIT;
+    if ((info->imageUsage & required_flags) != required_flags) {
+        PL_ERR(vk, "The swapchain doesn't support rendering and blitting!");
+        goto error;
+    }
+
     return true;
 
 error:
@@ -485,8 +499,10 @@ static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
     // It's invalid to trigger another swapchain recreation while there's more
     // than one swapchain already active, so we need to flush any pending
     // asynchronous swapchain release operations that may be ongoing
-    while (p->old_swapchain)
+    while (p->old_swapchain) {
+        vk_flush_commands(vk);
         vk_poll_commands(vk, UINT64_MAX);
+    }
 
     VkSwapchainCreateInfoKHR sinfo = p->protoInfo;
     sinfo.oldSwapchain = p->swapchain;
@@ -513,18 +529,24 @@ static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
 
     // Get the new swapchain images
     VK(vk->GetSwapchainImagesKHR(vk->dev, p->swapchain, &num_images, NULL));
-    vkimages = talloc_array(NULL, VkImage, num_images);
+    vkimages = talloc_zero_array(NULL, VkImage, num_images);
     VK(vk->GetSwapchainImagesKHR(vk->dev, p->swapchain, &num_images, vkimages));
 
+    for (int i = 0; i < num_images; i++)
+        VK_NAME(IMAGE, vkimages[i], "swapchain");
+
     // If needed, allocate some more semaphores
     while (num_images > p->num_sems) {
-        VkSemaphore sem_in, sem_out;
+        VkSemaphore sem_in = NULL, sem_out = NULL;
         static const VkSemaphoreCreateInfo seminfo = {
             .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO,
         };
         VK(vk->CreateSemaphore(vk->dev, &seminfo, VK_ALLOC, &sem_in));
         VK(vk->CreateSemaphore(vk->dev, &seminfo, VK_ALLOC, &sem_out));
 
+        VK_NAME(SEMAPHORE, sem_in, "swapchain in");
+        VK_NAME(SEMAPHORE, sem_out, "swapchain out");
+
         int idx = p->num_sems++;
         TARRAY_GROW(sw, p->sems_in, idx);
         TARRAY_GROW(sw, p->sems_out, idx);
@@ -540,8 +562,13 @@ static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
     TARRAY_GROW(sw, p->images, num_images);
     for (int i = 0; i < num_images; i++) {
         const VkExtent2D *ext = &sinfo.imageExtent;
-        p->images[i] = pl_vulkan_wrap(gpu, vkimages[i], ext->width, ext->height,
-                                      0, sinfo.imageFormat, sinfo.imageUsage);
+        p->images[i] = pl_vulkan_wrap(gpu, &(struct pl_vulkan_wrap_params) {
+            .image = vkimages[i],
+            .width = ext->width,
+            .height = ext->height,
+            .format = sinfo.imageFormat,
+            .usage = sinfo.imageUsage,
+        });
         if (!p->images[i])
             goto error;
     }
@@ -560,6 +587,9 @@ static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
     p->color_repr.bits.sample_depth = bits;
     p->color_repr.bits.color_depth = bits;
 
+    // Re-set the HDR metadata
+    pl_swapchain_hdr_metadata(sw, &p->hdr_metadata);
+
     talloc_free(vkimages);
     return true;
 
@@ -647,9 +677,13 @@ static bool vk_sw_submit_frame(const struct pl_swapchain *sw)
     VkSemaphore sem_out = p->sems_out[p->idx_sems++];
     p->idx_sems %= p->num_sems;
 
-    pl_vulkan_hold(gpu, p->images[p->last_imgidx],
-                   VK_IMAGE_LAYOUT_PRESENT_SRC_KHR,
-                   VK_ACCESS_MEMORY_READ_BIT, sem_out);
+    bool held = pl_vulkan_hold(gpu, p->images[p->last_imgidx],
+                               VK_IMAGE_LAYOUT_PRESENT_SRC_KHR,
+                               VK_ACCESS_MEMORY_READ_BIT, sem_out);
+    if (!held) {
+        PL_ERR(gpu, "Failed holding swapchain image for presentation");
+        return false;
+    }
 
     struct vk_cmd *cmd = pl_vk_steal_cmd(gpu);
     if (!cmd)
@@ -658,19 +692,15 @@ static bool vk_sw_submit_frame(const struct pl_swapchain *sw)
     p->frames_in_flight++;
     vk_cmd_callback(cmd, (vk_cb) present_cb, p, NULL);
 
-    vk_cmd_queue(vk, cmd);
+    vk_cmd_queue(vk, &cmd);
     if (!vk_flush_commands(vk))
         return false;
 
-    vk_rotate_queues(p->vk);
-
-    // Older nvidia drivers can spontaneously combust when submitting to the
-    // same queue as we're rendering from, in a multi-queue scenario. Safest
-    // option is to flush the commands first and then submit to the next queue.
-    // We can drop this hack in the future, I suppose.
     struct vk_cmdpool *pool = vk->pool_graphics;
     VkQueue queue = pool->queues[pool->idx_queues];
 
+    vk_rotate_queues(p->vk);
+
     VkPresentInfoKHR pinfo = {
         .sType = VK_STRUCTURE_TYPE_PRESENT_INFO_KHR,
         .waitSemaphoreCount = 1,
@@ -725,6 +755,41 @@ static bool vk_sw_resize(const struct pl_swapchain *sw, int *width, int *height)
     return ok;
 }
 
+static bool vk_sw_hdr_metadata(const struct pl_swapchain *sw,
+                               const struct pl_hdr_metadata *metadata)
+{
+    struct priv *p = TA_PRIV(sw);
+    struct vk_ctx *vk = p->vk;
+    if (!vk->SetHdrMetadataEXT)
+        return false;
+
+    if (!pl_color_transfer_is_hdr(p->color_space.transfer))
+        return false;
+
+    if (!p->swapchain && !vk_sw_recreate(sw, 0, 0))
+        return false;
+
+    if (!metadata)
+        return true;
+
+    // Remember the metadata so we can re-apply it after swapchain recreation
+    p->hdr_metadata = *metadata;
+
+    vk->SetHdrMetadataEXT(vk->dev, 1, &p->swapchain, &(VkHdrMetadataEXT) {
+        .sType = VK_STRUCTURE_TYPE_HDR_METADATA_EXT,
+        .displayPrimaryRed   = { metadata->prim.red.x,   metadata->prim.red.y },
+        .displayPrimaryGreen = { metadata->prim.green.x, metadata->prim.green.y },
+        .displayPrimaryBlue  = { metadata->prim.blue.x,  metadata->prim.blue.y },
+        .whitePoint = { metadata->prim.white.x, metadata->prim.white.y },
+        .maxLuminance = metadata->max_luma,
+        .minLuminance = metadata->min_luma,
+        .maxContentLightLevel = metadata->max_cll,
+        .maxFrameAverageLightLevel = metadata->max_fall,
+    });
+
+    return true;
+}
+
 bool pl_vulkan_swapchain_suboptimal(const struct pl_swapchain *sw)
 {
     struct priv *p = TA_PRIV(sw);
@@ -735,6 +800,7 @@ static struct pl_sw_fns vulkan_swapchain = {
     .destroy      = vk_sw_destroy,
     .latency      = vk_sw_latency,
     .resize       = vk_sw_resize,
+    .hdr_metadata = vk_sw_hdr_metadata,
     .start_frame  = vk_sw_start_frame,
     .submit_frame = vk_sw_submit_frame,
     .swap_buffers = vk_sw_swap_buffers,
